/*
Copyright 2016 Fixstars Corporation

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http ://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "path_aggregation.hpp"
#include "vertical_path_aggregation.hpp"
#include "horizontal_path_aggregation.hpp"
#include "oblique_path_aggregation.hpp"

namespace sgm {

template <size_t MAX_DISPARITY>
PathAggregation<MAX_DISPARITY>::PathAggregation()
	: m_cost_buffer()
{
	for(unsigned int i = 0; i < NUM_PATHS; ++i){
		cudaStreamCreate(&m_streams[i]);
		cudaEventCreate(&m_events[i]);
	}
	cudaEventCreate(&m_event);
}

template <size_t MAX_DISPARITY>
PathAggregation<MAX_DISPARITY>::~PathAggregation(){
	for(unsigned int i = 0; i < NUM_PATHS; ++i){
		cudaStreamSynchronize(m_streams[i]);
		cudaStreamDestroy(m_streams[i]);
		cudaEventDestroy(m_events[i]);
	}
	cudaEventDestroy(m_event);
}

template <size_t MAX_DISPARITY>
void PathAggregation<MAX_DISPARITY>::enqueue(
	const feature_type *left,
	const feature_type *right,
	int width,
	int height,
	unsigned int p1,
	const uint8_t *p2,
	int p2_pitch,
	const uint8_t* w,
	int w_pitch,
	cudaStream_t stream)
{
	const size_t buffer_size = width * height * MAX_DISPARITY * NUM_PATHS;
	if(m_cost_buffer.size() < buffer_size){
		m_cost_buffer = DeviceBuffer<cost_type>(buffer_size);
	}
	const size_t buffer_step = width * height * MAX_DISPARITY;
	//cudaStreamSynchronize(stream);
	cudaEventRecord(m_event, stream);

	for(unsigned int i = 0; i < NUM_PATHS; ++i){
		cudaStreamWaitEvent(m_streams[i], m_event, 0);
	}

	path_aggregation::enqueue_aggregate_up2down_path<MAX_DISPARITY>(
		m_cost_buffer.data() + 0 * buffer_step,
		left, right, width, height, p1, p2, p2_pitch, w, w_pitch, m_streams[0]);
	path_aggregation::enqueue_aggregate_down2up_path<MAX_DISPARITY>(
		m_cost_buffer.data() + 1 * buffer_step,
		left, right, width, height, p1, p2, p2_pitch, w, w_pitch, m_streams[1]);
	path_aggregation::enqueue_aggregate_left2right_path<MAX_DISPARITY>(
		m_cost_buffer.data() + 2 * buffer_step,
		left, right, width, height, p1, p2, p2_pitch, w, w_pitch, m_streams[2]);
	path_aggregation::enqueue_aggregate_right2left_path<MAX_DISPARITY>(
		m_cost_buffer.data() + 3 * buffer_step,
		left, right, width, height, p1, p2, p2_pitch, w, w_pitch, m_streams[3]);
	path_aggregation::enqueue_aggregate_upleft2downright_path<MAX_DISPARITY>(
		m_cost_buffer.data() + 4 * buffer_step,
		left, right, width, height, p1, p2, p2_pitch, w, w_pitch, m_streams[4]);
	path_aggregation::enqueue_aggregate_upright2downleft_path<MAX_DISPARITY>(
		m_cost_buffer.data() + 5 * buffer_step,
		left, right, width, height, p1, p2, p2_pitch, w, w_pitch, m_streams[5]);
	path_aggregation::enqueue_aggregate_downright2upleft_path<MAX_DISPARITY>(
		m_cost_buffer.data() + 6 * buffer_step,
		left, right, width, height, p1, p2, p2_pitch, w, w_pitch, m_streams[6]);
	path_aggregation::enqueue_aggregate_downleft2upright_path<MAX_DISPARITY>(
		m_cost_buffer.data() + 7 * buffer_step,
		left, right, width, height, p1, p2, p2_pitch, w, w_pitch, m_streams[7]);
	for(unsigned int i = 0; i < NUM_PATHS; ++i){
		cudaEventRecord(m_events[i], m_streams[i]);
		cudaStreamWaitEvent(stream, m_events[i], 0);
	}
}


template class PathAggregation< 64>;
template class PathAggregation<128>;
template class PathAggregation<256>;

}
